# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

# Use OSX cocoa/quartz to get glyph bitmaps

import numpy as np
from ctypes import byref, c_int32, c_byte

from ...ext.cocoapy import cf, ct, quartz, CFRange, CFSTR, CGGlyph, UniChar, \
    kCTFontFamilyNameAttribute, kCTFontBoldTrait, kCTFontItalicTrait, \
    kCTFontSymbolicTrait, kCTFontTraitsAttribute, kCTFontAttributeName, \
    kCGImageAlphaPremultipliedLast, kCFNumberSInt32Type, ObjCClass
from ._vispy_fonts import _vispy_fonts, _get_vispy_font_filename

_font_dict = {}


def _load_vispy_font(face, bold, italic):
    # http://stackoverflow.com/questions/2703085/
    # how-can-you-load-a-font-ttf-from-a-file-using-core-text
    fname = _get_vispy_font_filename(face, bold, italic)
    url = cf.CFURLCreateWithFileSystemPath(None, CFSTR(fname), 0, False)
    # data_provider = quartz.CGDataProviderCreateWithURL(url)
    # cg_font = quartz.CGFontCreateWithDataProvider(data_provider)
    # font = ct.CTFontCreateWithGraphicsFont(cg_font, 12., None, None)
    array = ct.CTFontManagerCreateFontDescriptorsFromURL(url)
    desc = cf.CFArrayGetValueAtIndex(array, 0)
    font = ct.CTFontCreateWithFontDescriptor(desc, 12., None)
    cf.CFRelease(array)
    cf.CFRelease(url)
    if not font:
        raise RuntimeError("Couldn't load font: %s" % face)
    key = '%s-%s-%s' % (face, bold, italic)
    _font_dict[key] = font
    return font


def _load_font(face, bold, italic):
    key = '%s-%s-%s' % (face, bold, italic)
    if key in _font_dict:
        return _font_dict[key]
    if face in _vispy_fonts:
        return _load_vispy_font(face, bold, italic)
    traits = 0
    traits |= kCTFontBoldTrait if bold else 0
    traits |= kCTFontItalicTrait if italic else 0

    # Create an attribute dictionary.
    args = [None, 0, cf.kCFTypeDictionaryKeyCallBacks,
            cf.kCFTypeDictionaryValueCallBacks]
    attributes = cf.CFDictionaryCreateMutable(*args)
    # Add family name to attributes.
    cfname = CFSTR(face)
    cf.CFDictionaryAddValue(attributes, kCTFontFamilyNameAttribute, cfname)
    cf.CFRelease(cfname)
    # Construct a CFNumber to represent the traits.
    itraits = c_int32(traits)
    sym_traits = cf.CFNumberCreate(None, kCFNumberSInt32Type, byref(itraits))
    if sym_traits:
        # Construct a dictionary to hold the traits values.
        args = [None, 0, cf.kCFTypeDictionaryKeyCallBacks,
                cf.kCFTypeDictionaryValueCallBacks]
        traits_dict = cf.CFDictionaryCreateMutable(*args)
        if traits_dict:
            # Add CFNumber traits to traits dictionary.
            cf.CFDictionaryAddValue(traits_dict, kCTFontSymbolicTrait,
                                    sym_traits)
            # Add traits dictionary to attributes.
            cf.CFDictionaryAddValue(attributes, kCTFontTraitsAttribute,
                                    traits_dict)
            cf.CFRelease(traits_dict)
        cf.CFRelease(sym_traits)
    # Create font descriptor with attributes.
    desc = ct.CTFontDescriptorCreateWithAttributes(attributes)
    cf.CFRelease(attributes)
    font = ct.CTFontCreateWithFontDescriptor(desc, 12., None)
    if not font:
        raise RuntimeError("Couldn't load font: %s" % face)
    _font_dict[key] = font
    return font


def _load_glyph(f, char, glyphs_dict):
    font = _load_font(f['face'], f['bold'], f['italic'])
    # resize loaded font
    args = [None, 0, cf.kCFTypeDictionaryKeyCallBacks,
            cf.kCFTypeDictionaryValueCallBacks]
    attributes = cf.CFDictionaryCreateMutable(*args)
    desc = ct.CTFontDescriptorCreateWithAttributes(attributes)
    cf.CFRelease(attributes)
    font = ct.CTFontCreateCopyWithAttributes(font, f['size'], None, desc)
    cf.CFRelease(desc)
    if not font:
        raise RuntimeError("Couldn't load font")
    # Create an attributed string using text and font.
    args = [None, 1, cf.kCFTypeDictionaryKeyCallBacks,
            cf.kCFTypeDictionaryValueCallBacks]
    attributes = cf.CFDictionaryCreateMutable(*args)
    cf.CFDictionaryAddValue(attributes, kCTFontAttributeName, font)
    string = cf.CFAttributedStringCreate(None, CFSTR(char), attributes)
    # Create a CTLine object to render the string.
    line = ct.CTLineCreateWithAttributedString(string)
    cf.CFRelease(string)
    cf.CFRelease(attributes)
    # Get a bounding rectangle for glyphs in string.
    chars = (UniChar * 1)(*map(ord, char))
    glyphs = (CGGlyph * 1)()
    ct.CTFontGetGlyphsForCharacters(font, chars, glyphs, 1)
    rect = ct.CTFontGetBoundingRectsForGlyphs(font, 0, glyphs, None, 1)
    # Get advance for all glyphs in string.
    advance = ct.CTFontGetAdvancesForGlyphs(font, 1, glyphs, None, 1)
    width = max(int(np.ceil(rect.size.width) + 1), 1)
    height = max(int(np.ceil(rect.size.height) + 1), 1)

    left = rect.origin.x
    baseline = -rect.origin.y
    top = height - baseline

    bits_per_component = 8
    bytes_per_row = 4*width
    color_space = quartz.CGColorSpaceCreateDeviceRGB()
    args = [None, width, height, bits_per_component, bytes_per_row,
            color_space, kCGImageAlphaPremultipliedLast]
    bitmap = quartz.CGBitmapContextCreate(*args)
    # Draw text to bitmap context.
    quartz.CGContextSetShouldAntialias(bitmap, True)
    quartz.CGContextSetTextPosition(bitmap, -left, baseline)
    ct.CTLineDraw(line, bitmap)
    cf.CFRelease(line)
    # Create an image to get the data out.
    image_ref = quartz.CGBitmapContextCreateImage(bitmap)
    assert quartz.CGImageGetBytesPerRow(image_ref) == bytes_per_row
    data_provider = quartz.CGImageGetDataProvider(image_ref)
    image_data = quartz.CGDataProviderCopyData(data_provider)
    buffer_size = cf.CFDataGetLength(image_data)
    assert buffer_size == width * height * 4
    buffer = (c_byte * buffer_size)()
    byte_range = CFRange(0, buffer_size)
    cf.CFDataGetBytes(image_data, byte_range, buffer)
    quartz.CGImageRelease(image_ref)
    quartz.CGDataProviderRelease(image_data)
    cf.CFRelease(bitmap)
    cf.CFRelease(color_space)

    # reshape bitmap (don't know why it's only alpha on OSX...)
    bitmap = np.array(buffer, np.ubyte)
    bitmap.shape = (height, width, 4)
    bitmap = bitmap[:, :, 3].copy()
    glyph = dict(char=char, offset=(left, top), bitmap=bitmap,
                 advance=advance, kerning={})
    glyphs_dict[char] = glyph
    # Generate kerning
    for other_char, other_glyph in glyphs_dict.items():
        glyph['kerning'][other_char] = (_get_k_p_a(font, other_char, char) -
                                        other_glyph['advance'])
        other_glyph['kerning'][char] = (_get_k_p_a(font, char, other_char) -
                                        glyph['advance'])
    cf.CFRelease(font)


def _get_k_p_a(font, left, right):
    """This actually calculates the kerning + advance"""
    # http://lists.apple.com/archives/coretext-dev/2010/Dec/msg00020.html
    # 1) set up a CTTypesetter
    chars = left + right
    args = [None, 1, cf.kCFTypeDictionaryKeyCallBacks,
            cf.kCFTypeDictionaryValueCallBacks]
    attributes = cf.CFDictionaryCreateMutable(*args)
    cf.CFDictionaryAddValue(attributes, kCTFontAttributeName, font)
    string = cf.CFAttributedStringCreate(None, CFSTR(chars), attributes)
    typesetter = ct.CTTypesetterCreateWithAttributedString(string)
    cf.CFRelease(string)
    cf.CFRelease(attributes)
    # 2) extract a CTLine from it
    range = CFRange(0, 1)
    line = ct.CTTypesetterCreateLine(typesetter, range)
    # 3) use CTLineGetOffsetForStringIndex to get the character positions
    offset = ct.CTLineGetOffsetForStringIndex(line, 1, None)
    cf.CFRelease(line)
    cf.CFRelease(typesetter)
    return offset


def _list_fonts():
    manager = ObjCClass('NSFontManager').sharedFontManager()
    avail = manager.availableFontFamilies()
    fonts = [avail.objectAtIndex_(ii).UTF8String().decode('utf-8')
             for ii in range(avail.count())]
    return fonts




from distutils.version import LooseVersion

import numpy as np

import glue
from glue.core import Data
from glue.core.component import Component
from ...common.tests.util import simple_session

from ..volume_viewer import VispyVolumeViewer
from ...extern.vispy.app import MouseEvent

GLUE_LT_08 = LooseVersion(glue.__version__) < LooseVersion('0.8')


def make_test_data():

    data = Data(label="Test Cube Data")

    np.random.seed(12345)

    for letter in 'abc':
        comp = Component(np.random.random((10, 10, 10)))
        data.add_component(comp, letter)

    # make sure one component key is primary
    data.add_component(Component(np.random.random((10, 10, 10))), 'PRIMARY')

    return data


def test_volumeviewer_toolbar():
    session = simple_session()
    v = VispyVolumeViewer(session)
    data = make_test_data()
    v.add_data(data)
    assert v.toolbar is not None

    toolbar = v.toolbar

    # test rotate tool
    toolbar.actions['vispy:rotate'].toggle()
    assert toolbar.active_tool.tool_id == 'vispy:rotate'
    # TODO: assert a mode here
    toolbar.actions['vispy:rotate'].toggle()
    assert toolbar.active_tool is None

    # test lasso selection tool
    toolbar.actions['vispy:lasso'].toggle()
    assert 'vispy:lasso' in toolbar.active_tool.tool_id
    lasso = toolbar.active_tool
    # event = QTest.mouseMove(viewer._vispy_widget)

    # TODO: add a real mouse move event so content in lasso.move() is called
    lasso.press(MouseEvent('mouse_press'))
    lasso.move(MouseEvent('mouse_move'))
    lasso.release(MouseEvent('mouse_release'))
    assert toolbar.tools['vispy:lasso'] == lasso

    # add point selection test
    # set to perspective mode
    toolbar.actions['volume3d:point'].toggle()
    assert 'volume3d:point' in toolbar.active_tool.tool_id
    point = toolbar.active_tool
    # event = QTest.mouseMove(viewer._vispy_widget)

    # TODO: add a real mouse move event so content in lasso.move() is called
    point.press(MouseEvent('mouse_press'))
    point.move(MouseEvent('mouse_move'))
    point.release(MouseEvent('mouse_release'))
    assert toolbar.tools['volume3d:point'] == point



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# This shim has been imported from Astropy.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

"""
Handle loading six package from system or from the bundled copy
"""

try:
    import importlib
except ImportError:
    importlib = None
    import imp
import io
import sys

from distutils.version import StrictVersion


_SIX_MIN_VERSION = StrictVersion('1.8.0')
_SIX_SEARCH_PATH = ['vispy.ext._bundled.six', 'six']


def _find_module(name, path=None):
    """
    Alternative to `imp.find_module` that can also search in subpackages.
    """

    parts = name.split('.')

    for part in parts:
        if path is not None:
            path = [path]

        fh, path, descr = imp.find_module(part, path)
        if fh is not None and part != parts[-1]:
            fh.close()

    return fh, path, descr


def _import_six(search_path=_SIX_SEARCH_PATH):
    for mod_name in search_path:
        if importlib is not None:
            try:
                six_mod = importlib.import_module(mod_name)
            except ImportError:
                continue
        else:
            try:
                mod_info = _find_module(mod_name)
            except ImportError:
                continue
            else:
                try:
                    # Using __name__ causes the import to effectively overwrite
                    # this shim.
                    six_mod = imp.load_module(__name__, *mod_info)
                finally:
                    if mod_info[0] is not None:
                        mod_info[0].close()

        try:
            if StrictVersion(six_mod.__version__) >= _SIX_MIN_VERSION:
                break
        except (AttributeError, ValueError):
            # Attribute error if the six module isn't what it should be and
            # doesn't have a .__version__; ValueError if the version string
            # exists but is somehow bogus/unparseable
            continue
    else:
        raise ImportError(
            "Vispy requires the 'six' module of minimum version {0}; "
            "normally this is bundled with the Vispy package so if you get "
            "this warning consult the packager of your Vispy "
            "distribution.".format(_SIX_MIN_VERSION))

    # Using importlib does not overwrite this shim, so do it ourselves.
    this_module = sys.modules[__name__]
    if not hasattr(this_module, '_importer'):
        # Copy all main six attributes.
        for name, value in six_mod.__dict__.items():
            if name.startswith('__'):
                continue
            this_module.__dict__[name] = value

        # Tell six's importer to accept this shim's name as its own.
        importer = six_mod._importer
        known_modules = list(importer.known_modules.items())
        for name, mod in known_modules:
            this_name = __name__ + name[len(mod_name):]
            importer.known_modules[this_name] = mod

        # Turn this shim into a package just like six does.
        this_module.__path__ = []  # required for PEP 302 and PEP 451
        this_module.__package__ = __name__  # see PEP 366
        if this_module.__dict__.get('__spec__') is not None:
            this_module.__spec__.submodule_search_locations = []  # PEP 451


_import_six()

if PY3:
    file_types = (io.TextIOWrapper, io.BufferedRandom)
else:
    file_types = (file, io.TextIOWrapper, io.BufferedRandom)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# glumpy is an OpenGL framework for the fast visualization of numpy arrays.
# Copyright (C) 2009-2011  Nicolas P. Rougier. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY NICOLAS P. ROUGIER ''AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL NICOLAS P. ROUGIER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of Nicolas P. Rougier.
# -----------------------------------------------------------------------------
'''
A filter is a shader that transform the current displayed texture. Since
shaders cannot be easily serialized within the GPU, they have to be well
structured on the python side such that we can possibly merge them into a
single source code for both vertex and fragment. Consequently, there is a
default code for both vertex and fragment with specific entry points such that
filter knows where to insert their specific code (declarations, functions and
call (or code) to be inserted in the main function).

Spatial interpolation filter classes for OpenGL textures.

Each filter generates a one-dimensional lookup table (weights value from 0 to
ceil(radius)) that is uploaded to video memory (as a 1d texture) and is then
read by the shader when necessary. It avoids computing weight values for each
pixel. Furthemore, each 2D-convolution filter is separable and can be computed
using 2 1D-convolution with same 1d-kernel (= the lookup table values).

Available filters:

  - Nearest  (radius 0.5)
  - Bilinear (radius 1.0)
  - Hanning (radius 1.0)
  - Hamming (radius 1.0)
  - Hermite (radius 1.0)
  - Kaiser (radius 1.0)
  - Quadric (radius 1.5)
  - Bicubic (radius 2.0)
  - CatRom (radius 2.0)
  - Mitchell (radius 2.0)
  - Spline16 (radius 2.0)
  - Spline36 (radius 4.0)
  - Gaussian (radius 2.0)
  - Bessel (radius 3.2383)
  - Sinc (radius 4.0)
  - Lanczos (radius 4.0)
  - Blackman (radius 4.0)


Note::

  Weights code has been translated from the antigrain geometry library
  available at http://www.antigrain.com/
'''


import math
import numpy as np


class SpatialFilter(object):
    ''' '''

    def __init__(self, radius=1.0):
        self.radius = radius

    def weight(self, x):
        '''
        Return filter weight for a distance x.

        :Parameters:
            ``x`` : 0 < float < ceil(self.radius)
                Distance to be used to compute weight.
        '''
        raise NotImplemented

    def kernel(self, size=4*512):
        radius = self.radius
        r = int(max(1.0, math.ceil(radius)))
        samples = size / r
        n = size  # r*samples
        kernel = np.zeros(n)
        X = np.linspace(0, r, n)
        for i in range(n):
            kernel[i] = self.weight(X[i])
        N = np.zeros(samples)
        for i in range(r):
            N += kernel[::+1][i*samples:(i+1)*samples]
            N += kernel[::-1][i*samples:(i+1)*samples]
        for i in range(r):
            kernel[i*samples:(i+1)*samples:+1] /= N
        return kernel

    def filter_code(self):

        n = int(math.ceil(self.radius))
        filter_1 = 'filter1D_radius%d' % n
        filter_2 = 'filter2D_radius%d' % n

        code = ''
        code += 'vec4\n'
        code += '%s( sampler2D kernel, float index, float x, ' % filter_1
        for i in range(2*n):
            if i == 2*n-1:
                code += 'vec4 c%d )\n' % i
            else:
                code += 'vec4 c%d, ' % i
        code += '{\n'
        code += '    float w, w_sum = 0.0;\n'
        code += '    vec4 r = vec4(0.0,0.0,0.0,0.0);\n'
        for i in range(n):
            code += '    w = unpack_interpolate(kernel, vec2(%f+(x/%.1f), index));\n' % (1.0 - (i + 1) / float(n), n)  # noqa
            code += '    w = w*kernel_scale + kernel_bias;\n'  # noqa
            # code += '   w_sum += w;'
            code += '    r += c%d * w;\n' % i
            code += '    w = unpack_interpolate(kernel, vec2(%f-(x/%.1f), index));\n' % ((i+1)/float(n), n)  # noqa
            code += '    w = w*kernel_scale + kernel_bias;\n'
            # code += '   w_sum += w;'
            code += '    r += c%d * w;\n' % (i + n)
        # code += '    return r/w_sum;\n'
        code += '    return r;\n'
        code += '}\n'
        code += "\n"
        code += 'vec4\n'
        code += '%s' % filter_2
        code += '(sampler2D texture, sampler2D kernel, float index, vec2 uv, vec2 pixel)\n'  # noqa
        code += '{\n'
        code += '    vec2 texel = uv/pixel - vec2(0.5, 0.5) ;\n'
        code += '    vec2 f = fract(texel);\n'
        code += '    texel = (texel-fract(texel) + vec2(0.001, 0.001)) * pixel;\n'  # noqa
        for i in range(2*n):
            code += '    vec4 t%d = %s(kernel, index, f.x,\n' % (i, filter_1)
            for j in range(2*n):
                x, y = (-n+1+j, -n+1+i)
                code += '        texture2D( texture, texel + vec2(%d, %d) * pixel),\n' % (x, y)  # noqa

            # Remove last trailing',' and close function call
            code = code[:-2] + ');\n'

        code += '    return %s(kernel, index, f.y, ' % filter_1
        for i in range(2*n):
            code += 't%d, ' % i

        # Remove last trailing',' and close function call
        code = code[:-2] + ');\n'
        code += '}\n'

        return code

    def call_code(self, index):
        code = ""
        n = int(math.ceil(self.radius))
        filter_1 = 'filter1D_radius%d' % n  # noqa
        filter_2 = 'filter2D_radius%d' % n

        code += 'vec4 %s(sampler2D texture, vec2 shape, vec2 uv)\n' % self.__class__.__name__  # noqa
        code += '{'
        code += ' return %s(texture, u_kernel, %f, uv, 1.0/shape); ' % (filter_2, index)  # noqa
        code += '}\n'
        return code


class Nearest(SpatialFilter):
    '''
    Nearest (=None) filter (radius = 0.5).

    Weight function::

      w(x) = 1

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=.5)

    def weight(self, x):
        return 1.0

    def _get_code(self):
        self.build_LUT()
        code = 'vec4\n'
        code += 'interpolate(sampler2D texture, sampler1D kernel, vec2 uv, vec2 pixel)\n'  # noqa
        code += '{\n   return texture2D(texture, uv);\n}\n'
        return code
    code = property(_get_code, doc='''filter functions code''')


class Bilinear(SpatialFilter):
    '''
    Bilinear filter (radius = 1.0).

    Weight function::

      w(x) = 1 - x

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=1.0)

    def weight(self, x):
        return 1.0 - x


class Hanning(SpatialFilter):
    '''
    Hanning filter (radius = 1.0).

    Weight function::

      w(x) = 0.5 + 0.5 * cos(pi * x)

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=1.0)

    def weight(self, x):
        return 0.5 + 0.5 * math.cos(math.pi * x)


class Hamming(SpatialFilter):
    '''
    Hamming filter (radius = 1.0).

    Weight function::

      w(x) = 0.54 + 0.46 * cos(pi * x)

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=1.0)

    def weight(self, x):
        return 0.54 + 0.46 * math.cos(math.pi * x)


class Hermite(SpatialFilter):
    ''' Hermite filter (radius = 1.0).

    Weight function::

      w(x) = (2*x-3)*x^2 + 1

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=1.0)

    def weight(self, x):
        return (2.0 * x - 3.0) * x * x + 1.0


class Quadric(SpatialFilter):
    '''
    Quadric filter (radius = 1.5).

    Weight function::

             |  0.0 ≤ x < 0.5: 0.75 - x*x
      w(x) = |  0.5 ≤ x < 1.5: 0.5 - (x-1.5)^2
             |  1.5 ≤ x      : 0

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=1.5)

    def weight(self, x):
        if x < 0.75:
            return 0.75 - x * x
        elif x < 1.5:
            t = x - 1.5
            return 0.5 * t * t
        else:
            return 0.0


class Bicubic(SpatialFilter):
    '''
    Bicubic filter (radius = 2.0).

    Weight function::

      w(x) = 1/6((x+2)^3 - 4*(x+1)^3 + 6*x^3 -4*(x-1)^3)
    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=2.0)

    def pow3(self, x):
        if x <= 0:
            return 0
        else:
            return x * x * x

    def weight(self, x):
        return (1.0/6.0) * (self.pow3(x + 2) -
                            4 * self.pow3(x + 1) +
                            6 * self.pow3(x) -
                            4 * self.pow3(x - 1))


class Kaiser(SpatialFilter):
    '''
    Kaiser filter (radius = 1.0).


    Weight function::

      w(x) = bessel_i0(a sqrt(1-x^2)* 1/bessel_i0(b)

    '''

    def __init__(self, b=6.33):
        self.a = b
        self.epsilon = 1e-12
        self.i0a = 1.0 / self.bessel_i0(b)
        SpatialFilter.__init__(self, radius=1.0)

    def bessel_i0(self, x):
        s = 1.0
        y = x * x / 4.0
        t = y
        i = 2
        while t > self.epsilon:
            s += t
            t *= float(y) / (i * i)
            i += 1
        return s

    def weight(self, x):
        if x > 1:
            return 0
        return self.bessel_i0(self.a * math.sqrt(1.0 - x * x)) * self.i0a


class CatRom(SpatialFilter):
    '''
    Catmull-Rom filter (radius = 2.0).

    Weight function::

             |  0 ≤ x < 1: 0.5*(2 + x^2*(-5+x*3))
      w(x) = |  1 ≤ x < 2: 0.5*(4 + x*(-8+x*(5-x)))
             |  2 ≤ x    : 0

    '''

    def __init__(self, size=256*8):
        SpatialFilter.__init__(self, radius=2.0)

    def weight(self, x):
        if x < 1.0:
            return 0.5 * (2.0 + x * x * (-5.0 + x * 3.0))
        elif x < 2.0:
            return 0.5 * (4.0 + x * (-8.0 + x * (5.0 - x)))
        else:
            return 0.0


class Mitchell(SpatialFilter):
    '''
    Mitchell-Netravali filter (radius = 2.0).

    Weight function::

             |  0 ≤ x < 1: p0 + x^2*(p2 + x*p3)
      w(x) = |  1 ≤ x < 2: q0 + x*(q1 + x*(q2 + x*q3))
             |  2 ≤ x    : 0

    '''

    def __init__(self, b=1.0/3.0, c=1.0/3.0):
        self.p0 = (6.0 - 2.0 * b) / 6.0
        self.p2 = (-18.0 + 12.0 * b + 6.0 * c) / 6.0
        self.p3 = (12.0 - 9.0 * b - 6.0 * c) / 6.0
        self.q0 = (8.0 * b + 24.0 * c) / 6.0
        self.q1 = (-12.0 * b - 48.0 * c) / 6.0
        self.q2 = (6.0 * b + 30.0 * c) / 6.0
        self.q3 = (-b - 6.0 * c) / 6.0
        SpatialFilter.__init__(self, radius=2.0)

    def weight(self, x):
        if x < 1.0:
            return self.p0 + x * x * (self.p2 + x * self.p3)
        elif x < 2.0:
            return self.q0 + x * (self.q1 + x * (self.q2 + x * self.q3))
        else:
            return 0.0


class Spline16(SpatialFilter):
    '''
    Spline16 filter (radius = 2.0).

    Weight function::

             |  0 ≤ x < 1: ((x-9/5)*x - 1/5)*x + 1
      w(x) = |
             |  1 ≤ x < 2: ((-1/3*(x-1) + 4/5)*(x-1) - 7/15 )*(x-1)

    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=2.0)

    def weight(self, x):
        if x < 1.0:
            return ((x - 9.0/5.0) * x - 1.0/5.0) * x + 1.0
        else:
            return ((-1.0/3.0 * (x-1) + 4.0/5.0) * (x-1) - 7.0/15.0) * (x-1)


class Spline36(SpatialFilter):
    '''
    Spline36 filter (radius = 3.0).

    Weight function::

             |  0 ≤ x < 1: ((13/11*x - 453/209)*x -3/209)*x +1
      w(x) = |  1 ≤ x < 2: ((-6/11*(x-1) - 270/209)*(x-1) -156/209)*(x-1)
             |  2 ≤ x < 3: (( 1/11*(x-2) - 45/209)*(x-2) + 26/209)*(x-2)
    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=3.0)

    def weight(self, x):
        if x < 1.0:
            return ((13.0/11.0 * x - 453.0/209.0) * x - 3.0/209.0) * x + 1.0
        elif x < 2.0:
            return ((-6.0/11.0 * (x-1) + 270.0/209.0) * (x-1) - 156.0 / 209.0) * (x-1)  # noqa
        else:
            return ((1.0 / 11.0 * (x-2) - 45.0/209.0) * (x - 2) + 26.0/209.0) * (x-2)  # noqa


class Gaussian(SpatialFilter):
    '''
    Gaussian filter (radius = 2.0).

    Weight function::

      w(x) = exp(-2x^2) * sqrt(2/pi)

    Note::

      This filter does not seem to be correct since:

        x = np.linspace(0, 1.0, 100 )
        f = weight
        z = f(x+1)+f(x)+f(1-x)+f(2-x)

        z should be 1 everywhere but it is not the case and it produces "grid
        effects".
    '''
    def __init__(self):
        SpatialFilter.__init__(self, radius=2.0)

    def weight(self, x):
        return math.exp(-2.0 * x * x) * math.sqrt(2.0 / math.pi)


class Bessel(SpatialFilter):
    '''
    Bessel filter (radius = 3.2383).
    '''

    def __init__(self):
        SpatialFilter.__init__(self, radius=3.2383)

    def besj(self, x, n):
        '''
        Function BESJ calculates Bessel function of first kind of order n
        Arguments:
            n - an integer (>=0), the order
            x - value at which the Bessel function is required
        --------------------
        C++ Mathematical Library
        Converted from equivalent FORTRAN library
        Converted by Gareth Walker for use by course 392 computational project
        All functions tested and yield the same results as the corresponding
        FORTRAN versions.

        If you have any problems using these functions please report them to
        M.Muldoon@UMIST.ac.uk

        Documentation available on the web
        http://www.ma.umist.ac.uk/mrm/Teaching/392/libs/392.html
        Version 1.0   8/98
        29 October, 1999
        --------------------
        Adapted for use in AGG library by
                    Andy Wilk (castor.vulgaris@gmail.com)
        Adapted for use in vispy library by
                    Nicolas P. Rougier (Nicolas.Rougier@inria.fr)
        -----------------------------------------------------------------------
        '''
        if n < 0:
            return 0.0

        d = 1e-6
        b = 0
        if math.fabs(x) <= d:
            if n != 0:
                return 0
            return 1

        b1 = 0  # b1 is the value from the previous iteration
        # Set up a starting order for recurrence
        m1 = int(math.fabs(x)) + 6
        if math.fabs(x) > 5:
            m1 = int(math.fabs(1.4 * x + 60 / x))

        m2 = int(n + 2 + math.fabs(x) / 4)
        if m1 > m2:
            m2 = m1

        # Apply recurrence down from curent max order
        while True:
            c3 = 0
            c2 = 1e-30
            c4 = 0
            m8 = 1
            if m2 / 2 * 2 == m2:
                m8 = -1

            imax = m2 - 2
            for i in range(1, imax+1):
                c6 = 2 * (m2 - i) * c2 / x - c3
                c3 = c2
                c2 = c6
                if m2 - i - 1 == n:
                    b = c6
                m8 = -1 * m8
                if m8 > 0:
                    c4 = c4 + 2 * c6

            c6 = 2 * c2 / x - c3
            if n == 0:
                b = c6
            c4 += c6
            b /= c4
            if math.fabs(b - b1) < d:
                return b
            b1 = b
            m2 += 3

    def weight(self, x):
        if x == 0.0:
            return math.pi/4.0
        else:
            return self.besj(math.pi * x, 1) / (2.0 * x)


class Sinc(SpatialFilter):
    '''
    Sinc filter (radius = 4.0).

    Weight function::


    '''

    def __init__(self, size=256, radius=4.0):
        SpatialFilter.__init__(self, radius=max(radius, 2.0))

    def weight(self, x):
        if x == 0.0:
            return 1.0
        x *= math.pi
        return (math.sin(x) / x)


class Lanczos(SpatialFilter):
    '''
    Lanczos filter (radius = 4.0).

    Weight function::


    '''

    def __init__(self, size=256, radius=4.0):
        SpatialFilter.__init__(self, radius=max(radius, 2.0))

    def weight(self, x):
        if x == 0.0:
            return 1.0
        elif x > self.radius:
            return 0.0
        x *= math.pi
        xr = x / self.radius
        return (math.sin(x) / x) * (math.sin(xr)/xr)


class Blackman(SpatialFilter):
    '''
    Blackman filter (radius = 4.0).
    '''

    def __init__(self, size=256, radius=4.0):
        SpatialFilter.__init__(self, radius=max(radius, 2.0))

    def weight(self, x):
        if x == 0.0:
            return 1.0
        elif x > self.radius:
            return 0.0
        x *= math.pi
        xr = x / self.radius
        return (math.sin(x) / x) * (0.42 + 0.5*math.cos(xr) + 0.08*math.cos(2*xr))  # noqa


# Generate kernels texture (16 x 1024)
filters = [Bilinear(), Hanning(),  Hamming(),  Hermite(),
           Kaiser(),   Quadric(),  Bicubic(),  CatRom(),
           Mitchell(), Spline16(), Spline36(), Gaussian(),
           Bessel(),   Sinc(),     Lanczos(),  Blackman()]

n = 1024
K = np.zeros((16, n))
for i, f in enumerate(filters):
    K[i] = f.kernel(n)

bias = K.min()
scale = K.max()-K.min()
K = (K-bias)/scale
np.save("spatial-filters.npy", K.astype(np.float32))

print("// ------------------------------------")
print("// Automatically generated, do not edit")
print("// ------------------------------------")
print("")
print("const float kernel_bias  = %f;" % bias)
print("const float kernel_scale = %f;" % scale)
print("const float kernel_size = %f;" % n)
print("const vec4 bits = vec4(1.0, 1.0/256.0, 1.0/(256.0*256.0), 1.0/(256.0*256.0*256.0));")  # noqa
print("uniform sampler2D u_kernel;")
print("")

code = 'float\n'
code += 'unpack_unit(vec4 rgba)\n'
code += '{\n'
code += '\t// return rgba.r;  // uncomment this for r32f debugging\n'
code += '\treturn dot(rgba, bits);\n'
code += '}\n'
print(code.expandtabs(4))

code = 'float\n'
code += 'unpack_ieee(vec4 rgba)\n'
code += '{\n'
code += '\t// return rgba.r;  // uncomment this for r32f debugging\n'
code += '\trgba.rgba = rgba.abgr * 255;\n'
code += '\tfloat sign = 1.0 - step(128.0,rgba[0])*2.0;\n'
code += '\tfloat exponent = 2.0 * mod(rgba[0],128.0) + ' \
        'step(128.0,rgba[1]) - 127.0;\n'
code += '\tfloat mantissa = mod(rgba[1],128.0)*65536.0 + rgba[2]*256.0 + ' \
        'rgba[3] + float(0x800000);\n'
code += '\treturn sign * exp2(exponent) * (mantissa * exp2(-23));\n'
code += '}\n'
print(code.expandtabs(4))

code = 'float\n'
code += 'unpack_interpolate(sampler2D kernel, vec2 uv)\n'
code += '{\n'
code += '\t// return texture2D(kernel, uv).r; ' \
        '//uncomment this for r32f debug without interpolation\n'
code += '\tfloat kpixel = 1/kernel_size;\n'
code += '\tfloat u = uv.x / kpixel;\n'
code += '\tfloat v = uv.y;\n'
code += '\tfloat uf = fract(u);\n'
code += '\tu = (u - uf) * kpixel;\n'
code += '\n'
code += '\tfloat d0 = unpack_unit(texture2D(kernel, vec2(u, v)));\n'
code += '\tfloat d1 = unpack_unit(texture2D(kernel, vec2(u + 1 * kpixel, v)));\n'  # noqa
code += '\treturn mix(d0, d1, uf);\n'
code += '}\n'
print(code.expandtabs(4))

F = SpatialFilter(1.0)
print(F.filter_code())
F = SpatialFilter(2.0)
print(F.filter_code())
F = SpatialFilter(3.0)
print(F.filter_code())
F = SpatialFilter(4.0)
print(F.filter_code())

# Generate filter functions
# Special case for nearest
print("""vec4 Nearest(sampler2D texture, vec2 shape, vec2 uv)""")
print("""{ return texture2D(texture,uv); }\n""")

for i, f in enumerate(filters):
    print(f.call_code((i+0.5)/16.0))



from __future__ import absolute_import, division, print_function

# Import vispy.gloo first when on Windows otherwise there are strange
# side-effects when PyQt4.Qt is imported first (which it now is in QtPy)
import sys
if sys.platform.startswith('win'):
    import glue_vispy_viewers.extern.vispy.gloo.gl  # noqa

from glue.utils.qt import get_qapp  # noqa

# The application has to always be referenced to avoid being shut down, so we
# keep a reference to it here
app = None


def pytest_configure(config):
    global app
    app = get_qapp()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

import numpy as np

from .mesh import MeshVisual
from ..ext.six import string_types


class HistogramVisual(MeshVisual):
    """Visual that calculates and displays a histogram of data

    Parameters
    ----------
    data : array-like
        Data to histogram. Currently only 1D data is supported.
    bins : int | array-like
        Number of bins, or bin edges.
    color : instance of Color
        Color of the histogram.
    orientation : {'h', 'v'}
        Orientation of the histogram.
    """
    def __init__(self, data, bins=10, color='w', orientation='h'):
        #   4-5
        #   | |
        # 1-2/7-8
        # |/| | |
        # 0-3-6-9
        data = np.asarray(data)
        if data.ndim != 1:
            raise ValueError('Only 1D data currently supported')
        if not isinstance(orientation, string_types) or \
                orientation not in ('h', 'v'):
            raise ValueError('orientation must be "h" or "v", not %s'
                             % (orientation,))
        X, Y = (0, 1) if orientation == 'h' else (1, 0)

        # do the histogramming
        data, bin_edges = np.histogram(data, bins)
        # construct our vertices
        rr = np.zeros((3 * len(bin_edges) - 2, 3), np.float32)
        rr[:, X] = np.repeat(bin_edges, 3)[1:-1]
        rr[1::3, Y] = data
        rr[2::3, Y] = data
        bin_edges.astype(np.float32)
        # and now our tris
        tris = np.zeros((2 * len(bin_edges) - 2, 3), np.uint32)
        offsets = 3 * np.arange(len(bin_edges) - 1,
                                dtype=np.uint32)[:, np.newaxis]
        tri_1 = np.array([0, 2, 1])
        tri_2 = np.array([2, 0, 3])
        tris[::2] = tri_1 + offsets
        tris[1::2] = tri_2 + offsets
        MeshVisual.__init__(self, rr, tris, color=color)



#
# Anti-Grain Geometry - Version 2.4
# Copyright (C) 2002-2005 Maxim Shemanarev (McSeem)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#
#   3. The name of the author may not be used to endorse or promote
#      products derived from this software without specific prior
#      written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
#
# Python translation by Nicolas P. Rougier
# Copyright (C) 2013 Nicolas P. Rougier. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY NICOLAS P. ROUGIER ''AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL NICOLAS P. ROUGIER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of Nicolas P. Rougier.
#

import math

import numpy as np


curve_distance_epsilon = 1e-30
curve_collinearity_epsilon = 1e-30
curve_angle_tolerance_epsilon = 0.01
curve_recursion_limit = 32
m_cusp_limit = 0.0
m_angle_tolerance = 10 * math.pi / 180.0
m_approximation_scale = 1.0
m_distance_tolerance_square = (0.5 / m_approximation_scale)**2


def calc_sq_distance(x1, y1, x2, y2):
    dx = x2 - x1
    dy = y2 - y1
    return dx * dx + dy * dy


def _curve3_recursive_bezier(points, x1, y1, x2, y2, x3, y3, level=0):
    if level > curve_recursion_limit:
        return

    # Calculate all the mid-points of the line segments
    x12 = (x1 + x2) / 2.
    y12 = (y1 + y2) / 2.
    x23 = (x2 + x3) / 2.
    y23 = (y2 + y3) / 2.
    x123 = (x12 + x23) / 2.
    y123 = (y12 + y23) / 2.

    dx = x3 - x1
    dy = y3 - y1
    d = math.fabs((x2 - x3) * dy - (y2 - y3) * dx)

    if d > curve_collinearity_epsilon:
        # Regular case
        if d * d <= m_distance_tolerance_square * (dx * dx + dy * dy):
            # If the curvature doesn't exceed the distance_tolerance value
            # we tend to finish subdivisions.
            if m_angle_tolerance < curve_angle_tolerance_epsilon:
                points.append((x123, y123))
                return

            # Angle & Cusp Condition
            da = math.fabs(
                math.atan2(y3 - y2, x3 - x2) - math.atan2(y2 - y1, x2 - x1))
            if da >= math.pi:
                da = 2 * math.pi - da

            if da < m_angle_tolerance:
                # Finally we can stop the recursion
                points.append((x123, y123))
                return
    else:
        # Collinear case
        da = dx * dx + dy * dy
        if da == 0:
            d = calc_sq_distance(x1, y1, x2, y2)
        else:
            d = ((x2 - x1) * dx + (y2 - y1) * dy) / da
            if d > 0 and d < 1:
                # Simple collinear case, 1---2---3, we can leave just two
                # endpoints
                return
            if(d <= 0):
                d = calc_sq_distance(x2, y2, x1, y1)
            elif d >= 1:
                d = calc_sq_distance(x2, y2, x3, y3)
            else:
                d = calc_sq_distance(x2, y2, x1 + d * dx, y1 + d * dy)

        if d < m_distance_tolerance_square:
            points.append((x2, y2))
            return

    # Continue subdivision
    _curve3_recursive_bezier(points, x1, y1, x12, y12, x123, y123, level + 1)
    _curve3_recursive_bezier(points, x123, y123, x23, y23, x3, y3, level + 1)


def _curve4_recursive_bezier(points, x1, y1, x2, y2, x3, y3, x4, y4, level=0):
    if level > curve_recursion_limit:
        return

    # Calculate all the mid-points of the line segments
    x12 = (x1 + x2) / 2.
    y12 = (y1 + y2) / 2.
    x23 = (x2 + x3) / 2.
    y23 = (y2 + y3) / 2.
    x34 = (x3 + x4) / 2.
    y34 = (y3 + y4) / 2.
    x123 = (x12 + x23) / 2.
    y123 = (y12 + y23) / 2.
    x234 = (x23 + x34) / 2.
    y234 = (y23 + y34) / 2.
    x1234 = (x123 + x234) / 2.
    y1234 = (y123 + y234) / 2.

    # Try to approximate the full cubic curve by a single straight line
    dx = x4 - x1
    dy = y4 - y1
    d2 = math.fabs(((x2 - x4) * dy - (y2 - y4) * dx))
    d3 = math.fabs(((x3 - x4) * dy - (y3 - y4) * dx))

    s = int((d2 > curve_collinearity_epsilon) << 1) + \
        int(d3 > curve_collinearity_epsilon)

    if s == 0:
        # All collinear OR p1==p4
        k = dx * dx + dy * dy
        if k == 0:
            d2 = calc_sq_distance(x1, y1, x2, y2)
            d3 = calc_sq_distance(x4, y4, x3, y3)

        else:
            k = 1. / k
            da1 = x2 - x1
            da2 = y2 - y1
            d2 = k * (da1 * dx + da2 * dy)
            da1 = x3 - x1
            da2 = y3 - y1
            d3 = k * (da1 * dx + da2 * dy)
            if d2 > 0 and d2 < 1 and d3 > 0 and d3 < 1:
                # Simple collinear case, 1---2---3---4
                # We can leave just two endpoints
                return

            if d2 <= 0:
                d2 = calc_sq_distance(x2, y2, x1, y1)
            elif d2 >= 1:
                d2 = calc_sq_distance(x2, y2, x4, y4)
            else:
                d2 = calc_sq_distance(x2, y2, x1 + d2 * dx, y1 + d2 * dy)

            if d3 <= 0:
                d3 = calc_sq_distance(x3, y3, x1, y1)
            elif d3 >= 1:
                d3 = calc_sq_distance(x3, y3, x4, y4)
            else:
                d3 = calc_sq_distance(x3, y3, x1 + d3 * dx, y1 + d3 * dy)

        if d2 > d3:
            if d2 < m_distance_tolerance_square:
                points.append((x2, y2))
                return
        else:
            if d3 < m_distance_tolerance_square:
                points.append((x3, y3))
                return

    elif s == 1:
        # p1,p2,p4 are collinear, p3 is significant
        if d3 * d3 <= m_distance_tolerance_square * (dx * dx + dy * dy):
            if m_angle_tolerance < curve_angle_tolerance_epsilon:
                points.append((x23, y23))
                return

            # Angle Condition
            da1 = math.fabs(
                math.atan2(y4 - y3, x4 - x3) - math.atan2(y3 - y2, x3 - x2))
            if da1 >= math.pi:
                da1 = 2 * math.pi - da1

            if da1 < m_angle_tolerance:
                points.extend([(x2, y2), (x3, y3)])
                return

            if m_cusp_limit != 0.0:
                if da1 > m_cusp_limit:
                    points.append((x3, y3))
                    return

    elif s == 2:
        # p1,p3,p4 are collinear, p2 is significant
        if d2 * d2 <= m_distance_tolerance_square * (dx * dx + dy * dy):
            if m_angle_tolerance < curve_angle_tolerance_epsilon:
                points.append((x23, y23))
                return

            # Angle Condition
            # ---------------
            da1 = math.fabs(
                math.atan2(y3 - y2, x3 - x2) - math.atan2(y2 - y1, x2 - x1))
            if da1 >= math.pi:
                da1 = 2 * math.pi - da1

            if da1 < m_angle_tolerance:
                points.extend([(x2, y2), (x3, y3)])
                return

            if m_cusp_limit != 0.0:
                if da1 > m_cusp_limit:
                    points.append((x2, y2))
                    return

    elif s == 3:
        # Regular case
        if (d2 + d3) * (d2 + d3) <= m_distance_tolerance_square * (
                dx * dx + dy * dy):
            # If the curvature doesn't exceed the distance_tolerance value
            # we tend to finish subdivisions.

            if m_angle_tolerance < curve_angle_tolerance_epsilon:
                points.append((x23, y23))
                return

            # Angle & Cusp Condition
            k = math.atan2(y3 - y2, x3 - x2)
            da1 = math.fabs(k - math.atan2(y2 - y1, x2 - x1))
            da2 = math.fabs(math.atan2(y4 - y3, x4 - x3) - k)
            if da1 >= math.pi:
                da1 = 2 * math.pi - da1
            if da2 >= math.pi:
                da2 = 2 * math.pi - da2

            if da1 + da2 < m_angle_tolerance:
                # Finally we can stop the recursion
                points.append((x23, y23))
                return

            if m_cusp_limit != 0.0:
                if da1 > m_cusp_limit:
                    points.append((x2, y2))
                    return

                if da2 > m_cusp_limit:
                    points.append((x3, y3))
                    return

    # Continue subdivision
    _curve4_recursive_bezier(
        points, x1, y1, x12, y12, x123, y123, x1234, y1234, level + 1)
    _curve4_recursive_bezier(
        points, x1234, y1234, x234, y234, x34, y34, x4, y4, level + 1)


def curve3_bezier(p1, p2, p3):
    """
    Generate the vertices for a quadratic Bezier curve.

    The vertices returned by this function can be passed to a LineVisual or
    ArrowVisual.

    Parameters
    ----------
    p1 : array
        2D coordinates of the start point
    p2 : array
        2D coordinates of the first curve point
    p3 : array
        2D coordinates of the end point

    Returns
    -------
    coords : list
        Vertices for the Bezier curve.

    See Also
    --------
    curve4_bezier

    Notes
    -----
    For more information about Bezier curves please refer to the `Wikipedia`_
    page.

    .. _Wikipedia: https://en.wikipedia.org/wiki/B%C3%A9zier_curve
    """
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    points = []
    _curve3_recursive_bezier(points, x1, y1, x2, y2, x3, y3)

    dx, dy = points[0][0] - x1, points[0][1] - y1
    if (dx * dx + dy * dy) > 1e-10:
        points.insert(0, (x1, y1))

    dx, dy = points[-1][0] - x3, points[-1][1] - y3
    if (dx * dx + dy * dy) > 1e-10:
        points.append((x3, y3))

    return np.array(points).reshape(len(points), 2)


def curve4_bezier(p1, p2, p3, p4):
    """
    Generate the vertices for a third order Bezier curve.

    The vertices returned by this function can be passed to a LineVisual or
    ArrowVisual.

    Parameters
    ----------
    p1 : array
        2D coordinates of the start point
    p2 : array
        2D coordinates of the first curve point
    p3 : array
        2D coordinates of the second curve point
    p4 : array
        2D coordinates of the end point

    Returns
    -------
    coords : list
        Vertices for the Bezier curve.

    See Also
    --------
    curve3_bezier

    Notes
    -----
    For more information about Bezier curves please refer to the `Wikipedia`_
    page.

    .. _Wikipedia: https://en.wikipedia.org/wiki/B%C3%A9zier_curve
    """
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4
    points = []
    _curve4_recursive_bezier(points, x1, y1, x2, y2, x3, y3, x4, y4)

    dx, dy = points[0][0] - x1, points[0][1] - y1
    if (dx * dx + dy * dy) > 1e-10:
        points.insert(0, (x1, y1))
    dx, dy = points[-1][0] - x4, points[-1][1] - y4
    if (dx * dx + dy * dy) > 1e-10:
        points.append((x4, y4))

    return np.array(points).reshape(len(points), 2)



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""
Provides functionality for composing shaders from multiple GLSL
code snippets.
"""

__all__ = ['ModularProgram', 'Function', 'MainFunction', 'Variable', 'Varying',
           'FunctionChain', 'Compiler', 'MultiProgram']

from .program import ModularProgram  # noqa
from .function import Function, MainFunction, FunctionChain  # noqa
from .function import StatementList  # noqa
from .variable import Variable, Varying  # noqa
from .compiler import Compiler  # noqa
from .multiprogram import MultiProgram  # noqa
